import asyncio
import sys

import uvicorn
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import RedirectResponse
from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from core.database.database import Database
from core.environment import Environment
from core.logger import Logger
from data.schemas.public.short_url import ShortUrl
from data.schemas.public.short_url_dao import shortUrlDao
from data.schemas.public.short_url_visit import ShortUrlVisit
from data.schemas.public.short_url_visit_dao import shortUrlVisitDao

logger = Logger(__name__)
templates = Jinja2Templates(directory="templates")


async def index(request: Request):
    return templates.TemplateResponse("404.html", {"request": request}, status_code=404)


async def handle_request(request: Request):
    path = request.path_params["path"]
    short_url = await shortUrlDao.find_single_by({ShortUrl.short_url: path})
    if short_url is None:
        return templates.TemplateResponse("404.html", {"request": request}, status_code=404)

    domains = Environment.get("DOMAINS", list[str], [])
    domain = await short_url.domain
    logger.debug(
        f"Domain: {domain.name if domain is not None else None}, request.host: {request.headers['host']}"
    )

    host = request.headers["host"]
    if ":" in host:
        host = host.split(":")[0]

    domain_strict_mode = Environment.get("DOMAIN_STRICT_MODE", bool, False)
    if domain is not None and (
            domain.name not in domains
            or (domain_strict_mode and not host.endswith(domain.name))
    ):
        return templates.TemplateResponse("404.html", {"request": request}, status_code=404)

    user_agent = request.headers.get("User-Agent", "").lower()

    if "wheregoes" in user_agent or "someothertool" in user_agent:
        return await _handle_short_url(request, path, short_url)

    if short_url.loading_screen:
        await _track_visit(request, short_url)

        return templates.TemplateResponse(
            "redirect.html",
            {"request": request, "key": short_url.short_url, "target_url": _get_redirect_url(short_url.target_url)},
        )

    return await _handle_short_url(request, path, short_url)


async def _handle_short_url(request: Request, path: str, short_url: ShortUrl):
    if path.startswith("api/"):
        path = path.replace("api/", "")

    await _track_visit(request, short_url)

    return RedirectResponse(_get_redirect_url(short_url.target_url))


async def _track_visit(request: Request, short_url: ShortUrl):
    try:
        await shortUrlVisitDao.create(
            ShortUrlVisit(0, short_url.id, request.headers.get("User-Agent"))
        )
    except Exception as e:
        logger.error(f"Failed to update short url {short_url.short_url} with error", e)


def _get_redirect_url(url: str) -> str:
    if not url.startswith("http://") and not url.startswith("https://"):
        url = f"http://{url}"

    return url


async def configure():
    Logger.set_level(Environment.get("LOG_LEVEL", str, "info"))
    Environment.set_environment(Environment.get("ENVIRONMENT", str, "production"))
    logger.info(f"Environment: {Environment.get_environment()}")

    await Database.startup_db()


routes = [
    Route("/", endpoint=index),
    Mount('/static', StaticFiles(directory='static'), name='static'),
    Route("/{path:path}", endpoint=handle_request),
]

app = Starlette(routes=routes, on_startup=[configure])


def main():
    if sys.platform == "win32":
        from asyncio import WindowsSelectorEventLoopPolicy

        asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())

    uvicorn.run(
        app,
        host="0.0.0.0",
        port=Environment.get("PORT", int, 5001),
        log_config=None,
    )


if __name__ == "__main__":
    main()